#!/usr/bin/perl
#
# Generate a mapping CSV of user accounts on the system for uidremap from
# the Active Directory using the algorithm used be idmap_rid
#
# Copyright (c) 2020 Jonathan A. Buzzard (jonathan at buzzard dot me dot uk)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# $Id: generate_mapping.pl,v 1.1 2020/06/17 08:57:36 jab Exp jab $
#
#

use strict;
use warnings;
use Net::LDAP;
use Term::ReadKey;

# adjust these to suit your site

# this is 500 on RHEL <7.0 and below and 1000 on just about everything else
my $MAX_SYS_PID = 1000;
# not present on all systems, traditionally a none system account with the highest possible UID
my $NFS_NOBODY = 65534;
# parameters for the idmap_rid algorithm
my $BASE_RID = 0;
my $LOW_RANGE_ID = 10000;
# details for your AD domain
my $domain = 'ds.strath.ac.uk';
my $base = 'DC=ds,DC=strath,DC=ac,DC=uk';



#
# Turn a binary SID into a string
#
sub BinarySID2String {
	my $binary = shift @_;

	my ( $revision_level, $sub_authority_count,
		$auth_id_16, $auth_id_32, @sub_authorities ) = unpack('C C n N V*', $binary);

	if ($sub_authority_count != scalar @sub_authorities) {
		warn "Invalid SID binary: $binary";
		return '';		
	}

	my $authority = ($auth_id_16*4294967296)+$auth_id_32;

	my $sid = join('-', 'S', $revision_level, $authority, @sub_authorities);

	return $sid;
}


#
# Prompt for a password
#
sub GetPassword {

	# tell the terminal not to show the typed chars
	Term::ReadKey::ReadMode('noecho');

	print "Password: ";
	my $password = Term::ReadKey::ReadLine(0);

	# reset the terminal to what it was previously doing
	Term::ReadKey::ReadMode('restore');

	# the one you typed didn't echo!
	print "\n";

	# get rid of that pesky line ending (and works on Windows)
	$password =~ s/\R\z//;

	return $password;
}


# get a username and password to connect to the AD with
my $username = getlogin();
if ($login eq 'root') {
	print "Username: ";
	$username = Term::ReadKey::ReadLine(0);
	$username =~ s/\R\z//;
}
my $login = "$username\@$domain";
my $password = GetPassword();

# open a connection to the Active Directory
my @attributes = ('objectSid');
my $ad = Net::LDAP->new($domain) or die "$@";
$ad->bind($login, password=>$password, version => 3);


my $file = 'mapping.csv';
open MAPPING, ">$file";

#
# loop through all the users on the system
#
while(my @e = getpwent) {
	# only consider user accounts
	if (($e[2] >= $MAX_SYS_PID) && ($e[2] != $NFS_NOBODY)) {

		my $username = $e[0];
		my $filter = "(&(objectClass=user)(objectCategory=person)" . "(|(SAMAccountName=$username)))";
		my $results = $ad->search(base=> "DC=ds,DC=strath,DC=ac,DC=uk", filter=>$filter, attrs => \@attributes);
		die $results->error if $results->code;

		my $count = $results->count;

		# skip while warning if more than one account in the AD for our username; this should never happen
		if ($count>1) {
			print STDERR "more that one username match found in Active Directory for $username\n";
			next;
		}

		# if account does not exist skip while warning
		if ($count==0) {
			print STDERR "no account for $username found in the Active Directory\n";
			print MAPPING "u,$e[2],-1\n";
			next;
		}

		# generate the mapping for the account and write to output
		foreach my $entry($results->entries) {
			my $sid = BinarySID2String($entry->get_value('objectSid'));
			$sid =~ /.*-(\d+)$/m;
			my $id = $1-$BASE_RID+$LOW_RANGE_ID;
			print MAPPING "u,$e[2],$id\n";
		}
	}
}

# cleanup our searching
endpwent();
close MAPPING;
$ad->unbind;

