/* uidremap.c -- mass change of uid/gid's in a directory hierarchy
 *
 * Copyright (c) 2020 Jonathan A. Buzzard (jonathan at buzzard dot me dot uk)
 *
 * $Log: uidremap.c,v $
 * Revision 1.4  2020/06/17 08:57:14  jab
 * updated license text to GPL v3.0
 *
 * Revision 1.3  2020/06/17 08:41:49  jab
 * fixed comparison so largest uid and gid get changed
 *
 * Revision 1.2  2020/06/16 16:35:36  jab
 * added changing of Posix ACL's using system libraries
 * general code tidyup
 *
 * Revision 1.1  2020/06/11 22:05:11  jab
 * Initial revision
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */

static const char rcsid[]="$Id: uidremap.c,v 1.4 2020/06/17 08:57:14 jab Exp jab $";

#define _XOPEN_SOURCE 500
#define _GNU_SOURCE
#include <ftw.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <acl/libacl.h>
 
/*
 * Globals, no less
 */
long total;
uid_t *uidmap;
gid_t *gidmap;
uid_t maxuid;
gid_t maxgid;
int acl;

#define VERSION_STRING "\
uidremap version 0.2\n\
Copyright (c) 2020 Jonathan A. Buzzard \074jonathan\100buzzard\056me\056uk\076\n"

#define USAGE_STRING "\
Usage: uidremap [-m <file>] [-a] <directory>\n\
Traverse a directory hierarchy changing the UID and GID of all files and directories\n\n\
  -m,--map      CSV file containing the UID and GID mappings\n\
  -a,--acl      process ACL's as well\n\
  -h,--help     display this help message\n\
  -v,--version  display version\n\
Report bugs to \074jonathan\100buzzard\056me\056uk\076\n"


/*
 * Read the mappings from a CSV file of the format <u|g>,<source id>,<destination id>
 */
void ReadMapping(char *mapfile)
{
	FILE* stream;
	char type;
	long i,from,to;
	
	maxuid = 0;
	maxgid = 0;

	/* read the mappings file to find the largest UID and GID */
	stream = fopen(mapfile, "r");
	if (stream==NULL) {
		fprintf(stderr, "uidremap: unable to open mapfile %s\n", mapfile);
		exit(0);
	}
	while (fscanf(stream, "%c,%ld,%ld\n", &type, &from, &to)!=EOF) {
		if ((type=='u') && ((uid_t) from > maxuid)) {
			maxuid = (uid_t) from;
		}
		if ((type=='g') && ((gid_t) from > maxgid)) {
			maxgid = (gid_t) from;
		}
	}
	fclose(stream);
	
	/* allocate the space for the mappings */
	uidmap = malloc(sizeof(uid_t)*maxuid);
	gidmap = malloc(sizeof(gid_t)*maxgid);

	/* initialize to -1 so the lchown function does nothing when no mapping provided */
	for (i=0;i<maxuid;i++) {
		uidmap[i] = -1;
	}
	for (i=0;i<maxgid;i++) {
		gidmap[i] = -1;
	}

	/* read the mapping file again and populate the map */
	stream = fopen(mapfile, "r");
	if (stream==NULL) {
		fprintf(stderr, "uidremap: unable to open mapfile %s\n", mapfile);
		exit(0);
	}
	while (fscanf(stream, "%c,%ld,%ld\n", &type, &from, &to)!=EOF) {
		if (type=='u') {
			uidmap[(long) from] = to;
		}
		if (type=='g') {
			gidmap[(long) from] = to;
		}
	}
	fclose(stream);
	
	return;
}


/*
 * Modify the ACL of a file for the new UID/GID mappings
 */
void ProcessACLS(const char *fpath, acl_type_t type)
{
	int i,changed;
	acl_t acl;
	acl_entry_t entry;
	acl_tag_t tag;
	uid_t *uid,newuid;
	gid_t *gid,newgid;


	acl = acl_get_file(fpath, type);
	if (acl==NULL) {
		fprintf(stderr, "uidremap: unable to get access ACL for %s\n", fpath);
	} else {
		/* walk through each entry in this ACL */
		changed = 0;
		for (i=ACL_FIRST_ENTRY;;i=ACL_NEXT_ENTRY) {
		        if (acl_get_entry(acl, i, &entry)!=1) {
				/* exit loop on error or no more entries */
				break;
			}

		        if (acl_get_tag_type(entry, &tag) == -1) {
				fprintf(stderr, "acl_get_tag_type %s\n", fpath);
				break;
			}

			/* modify user ACL's */
			if (tag==ACL_USER) {
				uid = acl_get_qualifier(entry);
				if (uid == NULL) {
					fprintf(stderr, "acl_get_qualifier %s\n", fpath);
					break;
				}
				newuid = -1 ;
				if (*uid<=maxuid) {
					newuid = uidmap[(long) *uid];
				}
				if (newuid!=-1) {
					if (acl_set_qualifier(entry, &newuid)==-1) {
						fprintf(stderr, "acl_set_qualifier %s\n", fpath);
						break;
					}
					changed++;
				}
			}

			/* modify group ACL's */
			if (tag==ACL_GROUP) {
				gid = acl_get_qualifier(entry);
				if (gid == NULL) {
					fprintf(stderr, "acl_get_qualifier %s\n", fpath);
					break;
				}
				newgid = -1 ;
				if (*gid<=maxgid) {
					newgid = gidmap[(long) *gid];
				}
				if (newgid!=-1) {
					if (acl_set_qualifier(entry, &newgid)==-1) {
						fprintf(stderr, "acl_set_qualifier %s\n", fpath);
						break;
					}
					changed++;
				}
			}
		}

		/* if we modified an ACL write it back out */
		if (changed>0) {
			if (acl_set_file(fpath, type, acl)==-1) {
				fprintf(stderr, "uidremap: unable to set updated ACL on %s\n", fpath);
			}
		}
		acl_free(acl);
	}

	return;
}


/*
 * Process each file/directory in turn from the filesystem walk
 */
static int ProcessFiles(const char *fpath, const struct stat *sb, int tflag, struct FTW *ftwbuf)
{
	uid_t uid;
	gid_t gid;
	int error;

	/* warn if the directory cannot be read */
	if (tflag==FTW_DNR) {
		fprintf(stderr, "cannot read directory %s\n", fpath);
		return FTW_CONTINUE;
	}

	/*  get the mapping for the file making sure the UID/GID is inside our map range */
	if (sb->st_uid<=maxuid) {
		uid = uidmap[(long) sb->st_uid];
	} else {
		uid = -1;
	}
	if (sb->st_gid<=maxgid) {
		gid = gidmap[(long) sb->st_gid];
	} else {
		gid = -1;
	}

#ifdef DEBUG
	error = 0;
	printf("%s %d,%d:%d,%d\n", fpath, (int) sb->st_uid, (int) uid, (int) sb->st_gid, (int) gid);
#else
	error = lchown(fpath, uid, gid);
#endif

	if (error!=0) {
		fprintf(stderr, "uidremap: error changing ownership on %s\n", fpath);
	} else {
		total++;
	}

	/* process ACL's if requried and file is not a symlink and actually has an ACL */
	if ((acl) && (!S_ISLNK(sb->st_mode))) {
		/* an optimized Linux extension to determine if the file has an extended ACL  */
		error = acl_extended_file_nofollow(fpath);
		if (error==-1) {
			fprintf(stderr, "uidremap: error determining if ACL exits on %s\n", fpath);
		} else if (error==1) {
			/* deal with access ACL's first */
			ProcessACLS(fpath, ACL_TYPE_ACCESS);
			/* if this is a directory deal with any default ACL's too */
			if (S_ISDIR(sb->st_mode)) {
				ProcessACLS(fpath, ACL_TYPE_DEFAULT);
			}
		}
	}

	/* to tell nftw() to continue */
	return FTW_CONTINUE;
}


/*
 * Entry point.
 */
int main(int argc, char *argv[])
{
	int i,map;
	char mapfile[PATH_MAX];
	char hierarchy[PATH_MAX];


	/* this program *must* be run as root (make no sense otherwise) */
	if (getuid()) {
		fprintf(stderr, "uidremap: must be run as root.\n" );
		exit(-1);
	}

	acl = 0;
	map = 0;
	for (i=1;i<argc;i++) {
		if ((!strcmp(argv[i], "-h")) || (!strcmp(argv[i], "--help"))) {
			fprintf(stdout, USAGE_STRING);
			exit(0);
		} else if ((!strcmp(argv[i], "-v")) || (!strcmp(argv[i], "--version"))) {
			fprintf(stdout, VERSION_STRING);
			exit(0);
		} else if ((!strcmp(argv[i], "-a")) || (!strcmp(argv[i], "--acl"))) {
			acl = 1;
		} else if ((!strcmp(argv[i], "-m")) || (!strcmp(argv[i], "--map"))) {
			strncat(mapfile, argv[++i], PATH_MAX-1);
			map = 1;
		} else {
			strncat(hierarchy, argv[i],PATH_MAX-1);
		}
	}

	if (map==0) {
		fprintf(stderr, "uidremap: no map file supplied\n");
		exit(0);
	}

	if (strnlen(hierarchy, PATH_MAX)==0) {
		fprintf(stderr, "uidremap: no directory hierachy supplied\n");
		exit(0);
	} else if ((strnlen(hierarchy, PATH_MAX)==1) && (hierarchy[0]=='.')) {
		/* . supplied so get the current working directory */
		if (getcwd(hierarchy, PATH_MAX)==NULL) {
			fprintf(stderr, "uidremap: error getting current working directory\n");
			exit(0);
		}
	}

	/* read the mapping file from disk and populate the mapping tables */
	ReadMapping(mapfile);

	/* start the file tree walk */
	total = 0;
	nftw(hierarchy, ProcessFiles, 50, FTW_MOUNT | FTW_ACTIONRETVAL | FTW_PHYS);

	printf("TOTAL NUMBER OF FILE MODIFIED = %ld\n", total);

	return 0;
}
